# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_util.ipynb.

# %% auto 0
__all__ = ['nextv', 'enumdict', 'evalue', 'missing', 'get_torch_module']

# %% ../nbs/02_util.ipynb 6
from abc import ABCMeta
from pathlib import Path
from enum import Enum, EnumMeta, _EnumDict, member, StrEnum, auto, IntEnum
from importlib import import_module

# %% ../nbs/02_util.ipynb 8
from types import ModuleType, FunctionType
from typing import (
    Any, Type, Self, Union, Tuple, Iterable, TypedDict, TypeAlias, 
    Callable, Generator, Optional, ParamSpec, TypeVar, TypeGuard, Literal
)

# %% ../nbs/02_util.ipynb 10
#| export


# %% ../nbs/02_util.ipynb 12
from atyp import Layer, BoolQ
from nchr import NIL
from nlit import VALUE
from chck import isnone, notnone

# %% ../nbs/02_util.ipynb 14
from .atyp import T

# %% ../nbs/02_util.ipynb 16
def nextv(name: str, start: int, count: int, last_values: list[str]):
    return name


def enumdict(name: str) -> _EnumDict:
    newdict = _EnumDict()
    newdict._cls_name = name
    newdict._generate_next_value = nextv
    # try: newdict._generate_next_value = StrEnum._generate_next_value_
    # except: newdict._generate_next_value = Enum._generate_next_value_
    return newdict

def evalue(e: Enum): 
    return getattr(e, VALUE, e)

# %% ../nbs/02_util.ipynb 17
def missing(enm: Type[T], val: str | T, default = None) -> T:
    '''Method called for missing members during lookup.

    Parameters
    ----------
    cls : Enum
        The enumeration class.
    
    val : str
        The value to check for in the enum members.
    

    Returns
    -------
    Enum member or default
        The matched enum member or the default value.
    '''
    # for mem in enm:
    #     mal: member = getattr(mem, VALUE, '')
    #     if str(mal).casefold() == str(val).casefold(): return mem
    # if hasattr(enm, '_default_'): return enm._default_() or default
    # return default

    for mem in enm:
        mal = getattr(mem, 'value', '')        
        if str(mal).casefold() == str(val).casefold(): return mem
    if hasattr(enm, '_default_'): 
        edef = enm._default_()
        try: return enm(edef)
        except: ...
        try: return enm(default)
        except: return default
    return default

# %% ../nbs/02_util.ipynb 18
def get_torch_module(
    obj, *args, __module: ModuleType = 'torch.nn', 
    __enumkey: BoolQ = None, __enumval: BoolQ = None, __suffix: str = NIL,
    **kwargs
) -> Layer:
    '''
    Fetch a torch module based on an enumeration value or its name.

    Parameters
    ----------
    obj : Enum
        An enumeration instance or class.
    args : tuple
        Positional arguments to pass to the module constructor.
    __module : ModuleType, optional
        The torch module to fetch from, default is `torch.nn`.
    __enumkey : BoolQ, optional
        If True, use the enum's name for fetching, otherwise use the value.
    __enumval : BoolQ, optional
        If True, use the enum's value for fetching, otherwise use the name.
    __suffix : str, optional
        Suffix to append to the enum value or name.
    kwargs : dict
        Keyword arguments to pass to the module constructor.

    Returns
    -------
    Layer
        The instantiated torch module.
    '''
    # ensure that only one of __usekey and __useval is True
    usekey, useval = None, None
    if notnone(__enumkey): usekey = __enumkey
    if notnone(__enumval): useval = __enumval
    if isnone(usekey) and isnone(useval): usekey = True
    
    usekey = not useval if notnone(useval) else usekey
    useval = not usekey if notnone(usekey) else useval
    
    if issubclass(type(obj), Enum):
        val = getattr(obj, 'name' if usekey else 'value', obj.value)
    else: val = obj
        
    val = f'{val}{__suffix}'
    
    if __module is None: __module = 'torch.nn'
    if isinstance(__module, str): __module = import_module(__module)
    mod = getattr(__module, val)
    if args or kwargs: mod = mod(*args, **kwargs)
    return mod

# %% ../nbs/02_util.ipynb 20
#| export
